﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Persistence.Default.Xml;

namespace HeuristicLab.OptimizationExpertSystem {
  [Item("Expert-System", "Currently in experimental phase, an expert system that makes algorithm suggestions based on fitness landscape analysis features and an optimization knowledge base.")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 119)]
  public sealed class ExpertSystem : NamedItem, IStorableContent, INotifyPropertyChanged {

    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return VSImageLibrary.Library; }
    }

    [Storable]
    private int maximumEvaluations;
    public int MaximumEvaluations {
      get { return maximumEvaluations; }
      set {
        if (maximumEvaluations == value) return;
        maximumEvaluations = value;
        OnPropertyChanged("MaximumEvaluations");
        UpdateSuggestions();
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
    }

    [Storable]
    private ItemList<IAlgorithm> algorithmInstances;
    public ItemList<IAlgorithm> AlgorithmInstances {
      get { return algorithmInstances; }
      set {
        if (algorithmInstances == value) return;
        algorithmInstances = value;
        OnPropertyChanged("AlgorithmInstances");
        // TODO: Attach event handlers
      }
    }

    [Storable]
    private ISingleObjectiveHeuristicOptimizationProblem problem;
    public ISingleObjectiveHeuristicOptimizationProblem Problem {
      get { return problem; }
      set {
        if (problem == value) return;
        problem = value;
        OnPropertyChanged("Problem");
        UpdateSuggestions();
      }
    }

    [Storable]
    private ItemList<IAlgorithm> suggestedInstances;
    private ReadOnlyItemList<IAlgorithm> readOnlySuggestedInstances;
    public ReadOnlyItemList<IAlgorithm> SuggestedInstances {
      get { return readOnlySuggestedInstances; }
    }

    [Storable]
    private RunCollection problemInstances;
    public RunCollection ProblemInstances {
      get { return problemInstances; }
      set {
        if (problemInstances == value) return;
        problemInstances = value;
        OnPropertyChanged("ProblemInstances");
      }
    }

    private bool Maximization {
      get { return Problem != null && ((IValueParameter<BoolValue>)Problem.MaximizationParameter).Value.Value; }
    }

    [StorableConstructor]
    private ExpertSystem(bool deserializing) : base(deserializing) { }
    private ExpertSystem(ExpertSystem original, Cloner cloner)
      : base(original, cloner) {
      runs = cloner.Clone(original.runs);
      algorithmInstances = cloner.Clone(original.algorithmInstances);
      problem = cloner.Clone(original.problem);
      suggestedInstances = cloner.Clone(original.suggestedInstances);
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = cloner.Clone(original.problemInstances);
      RegisterEventHandlers();
    }
    public ExpertSystem() {
      Name = ItemName;
      Description = ItemDescription;
      runs = new RunCollection();
      algorithmInstances = new ItemList<IAlgorithm>();
      suggestedInstances = new ItemList<IAlgorithm>();
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = new RunCollection();
      RegisterEventHandlers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExpertSystem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      runs.CollectionReset += InformationChanged;
      runs.ItemsAdded += InformationChanged;
      runs.ItemsRemoved += InformationChanged;
      runs.Reset += InformationChanged;
      runs.UpdateOfRunsInProgressChanged += InformationChanged;
      algorithmInstances.CollectionReset += InformationChanged;
      algorithmInstances.ItemsAdded += InformationChanged;
      algorithmInstances.ItemsRemoved += InformationChanged;
    }

    private void InformationChanged(object sender, EventArgs e) {
      var runCollection = sender as RunCollection;
      if (runCollection != null && runCollection.UpdateOfRunsInProgress) return;
      UpdateSuggestions();
    }

    public void UpdateInstanceMap() {
      var flaValues = ProblemInstances.ResultNames
                      .Where(x => x.StartsWith("Characteristic.")
                        && ProblemInstances.All(y => y.Results.ContainsKey(x) && y.Results[x] is DoubleValue))
                      .Select(x => x).ToList();
      var ds = new double[ProblemInstances.Count, flaValues.Count];
      var instanceCounter = 0;
      foreach (var instance in ProblemInstances) {
        for (var feature = 0; feature < flaValues.Count; feature++)
          ds[instanceCounter, feature] = ((DoubleValue)instance.Results[flaValues[feature]]).Value;
        instanceCounter++;
      }

      int info;
      double[] s2;
      double[,] v;
      alglib.pcabuildbasis(ds, instanceCounter, flaValues.Count, out info, out s2, out v);

      var algInstRunDict = AlgorithmInstances.SelectMany(x => x.Runs)
                                             .Where(x => x.Parameters.ContainsKey("Problem Name") && x.Parameters["Problem Name"] is StringValue)
                                             .GroupBy(x => ((StringValue)x.Parameters["Problem Name"]).Value)
                                             .ToDictionary(x => x.Key, x => x.GroupBy(y => ((StringValue)y.Parameters["Algorithm Name"]).Value)
                                                                             .ToDictionary(y => y.Key, y => y.ToList()));
      ProblemInstances.UpdateOfRunsInProgress = true;
      try {
        instanceCounter = 0;
        foreach (var instance in ProblemInstances) {
          IItem probNameParam;
          if (!instance.Parameters.TryGetValue("Problem Name", out probNameParam)) continue;
          var probInstanceName = ((StringValue)probNameParam).Value;

          double x = 0, y = 0;
          for (var feature = 0; feature < flaValues.Count; feature++) {
            x += ds[instanceCounter, feature] * v[feature, 0];
            y += ds[instanceCounter, feature] * v[feature, 1];
          }
          IItem item;
          if (instance.Results.TryGetValue("Projection.PCA1", out item)) {
            ((DoubleValue)item).Value = x;
          } else instance.Results.Add("Projection.PCA1", new DoubleValue(x));
          if (instance.Results.TryGetValue("Projection.PCA2", out item)) {
            ((DoubleValue)item).Value = y;
          } else instance.Results.Add("Projection.PCA2", new DoubleValue(y));

          instanceCounter++;

          var bkQuality = ((DoubleValue)instance.Parameters["BestKnownQuality"]).Value;
          var maximization = ((BoolValue)instance.Parameters["Maximization"]).Value;

          if (!algInstRunDict.ContainsKey(probInstanceName)) continue;
          foreach (var kvp in algInstRunDict[probInstanceName]) {
            var algInstanceName = kvp.Key;
            // TODO: Things needs to be configurable here (table name, targets)
            foreach (var target in new[] { 1, 1.01, 1.05, 1.1, 1.2, 1.5 }) {
              var result = ExpectedRuntimeHelper.CalculateErt(kvp.Value, "QualityPerEvaluations", bkQuality * target, maximization);
              var resultName = algInstanceName + "@" + ((target - 1) * 100) + "%";
              if (instance.Results.TryGetValue(resultName, out item)) {
                ((DoubleValue)item).Value = Math.Log10(result.ExpectedRuntime);
              } else instance.Results.Add(resultName, new DoubleValue(Math.Log10(result.ExpectedRuntime)));
            }
          }
        }
      } finally { ProblemInstances.UpdateOfRunsInProgress = false; }
    }

    private static readonly HashSet<string> InterestingValueNames = new HashSet<string>() {
      "QualityPerEvaluations", "Problem Name", "Problem Type", "Algorithm Name", "Algorithm Type", "Maximization"
    };

    public async void UpdateKnowledgeBaseAsync(IProgress progress) {
      progress.Start("Updating Knowledge Base from OKB");
      await Task.Factory.StartNew(() => { DoUpdateKnowledgeBase(progress); }, TaskCreationOptions.LongRunning);
    }

    public void UpdateKnowledgeBase() {
      DoUpdateKnowledgeBase(new Progress(string.Empty, ProgressState.Started));
    }

    private void DoUpdateKnowledgeBase(IProgress progress) {
      var queryClient = Clients.OKB.Query.QueryClient.Instance;
      var adminClient = Clients.OKB.Administration.AdministrationClient.Instance;
      try {
        progress.Status = "Refreshing query client...";
        queryClient.Refresh();
        var interestingValues = queryClient.ValueNames.Where(x => InterestingValueNames.Contains(x.Name)).ToList();

        var problemTypeFilter = (Clients.OKB.Query.StringComparisonAvailableValuesFilter)queryClient.Filters.Single(x => x.Label == "Problem Data Type Name");
        problemTypeFilter.Value = Problem.GetType().Name;

        progress.Status = "Obtaining number of runs...";
        var count = queryClient.GetNumberOfRuns(problemTypeFilter);
        if (count == 0) return;

        var runIds = queryClient.GetRunIds(problemTypeFilter).ToList();
        adminClient.Refresh();
        var i = 0;
        var conversions = new List<Task>();
        var runGroup = new Dictionary<Tuple<long, string>, List<IRun>>();
        while (i < count) {
          var nextIds = runIds.Skip(i).Take(500).ToList();
          progress.Status = string.Format("Downloading runs {0} to {1} of {2}...", i, i + nextIds.Count, count);
          var okbRuns = queryClient.GetRunsWithValues(nextIds, true, interestingValues);
          conversions.Add(Task.Factory.StartNew(() => {
            var hlRuns = okbRuns.AsParallel().Select(x => new { Id = x.Algorithm.Id, Name = x.Algorithm.Name, Run = queryClient.ConvertToOptimizationRun(x) })
                                .GroupBy(x => Tuple.Create(x.Id, x.Name)).ToList();
            lock (runGroup) {
              foreach (var hr in hlRuns) {
                List<IRun> runList;
                if (!runGroup.TryGetValue(hr.Key, out runList)) {
                  runList = new List<IRun>();
                  runGroup[hr.Key] = runList;
                }
                runList.AddRange(hr.Select(x => x.Run));
              }
            }
          }));
          i += nextIds.Count;
          progress.ProgressValue = 0.8 * (i / (double)count);
        }
        Task.WaitAll(conversions.ToArray());
        var list = new ItemList<IAlgorithm>();
        i = 0;
        foreach (var rGroup in runGroup) {
          progress.Status = string.Format("Downloading algorithm {0}...", rGroup.Key.Item2);
          var data = Clients.OKB.Administration.AdministrationClient.GetAlgorithmData(rGroup.Key.Item1);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var alg = (IAlgorithm)XmlParser.Deserialize<IContent>(stream);
                alg.Runs.AddRange(rGroup.Value);
                list.Add(alg);
              } catch (Exception) { }
              stream.Close();
            }
          }
          i++;
          progress.ProgressValue = 0.8 + 0.2 * (i / (double)runGroup.Count);
        }
        AlgorithmInstances = list;
      } finally { progress.Finish(); }
    }

    private void UpdateSuggestions() {
      if (Problem == null) return;
      var instances = new SortedList<double, IAlgorithm>();
      foreach (var instance in algorithmInstances) {
        var relevantRuns = instance.Runs.Where(x => ((StringValue)x.Parameters["Problem Type"]).Value == Problem.GetType().Name);
        var avgQuality = 0.0;
        var counter = 0;
        foreach (var run in relevantRuns) {
          var performanceGraph = ((IndexedDataTable<double>)run.Results["QualityPerEvaluations"]);
          try {
            avgQuality += performanceGraph.Rows.First().Values.TakeWhile(x => x.Item1 < MaximumEvaluations).Last().Item2;
            counter++;
          } catch { continue; }
        }
        avgQuality /= counter;
        instances.Add(avgQuality, instance);
      }

      suggestedInstances.Clear();
      var instanceLadder = instances.Select(x => x.Value);
      suggestedInstances.AddRange(Maximization ? instanceLadder.Reverse() : instanceLadder);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string propertyName) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }
  }
}
